package net.christopherschultz.util.zipcode;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;
import java.io.Serializable;

/**
 * <p>A US-postal code.</p>
 *
 * <p>Although the latitude and longitude are usually for the post office that
 * serves each zip code, the value doesn't matter a whole lot. Zip
 * codes are usually fairly small, so any latitude/longitude value within
 * the zip code will work for most purposes.</p>
 *
 * @author Chris Schultz
 * @version $Revision$ $Date$
 */
public class ZipCode
    implements Comparable, Serializable
{
    /**
     * <p>The radius of the Earth, in miles.</p>
     */
    private static final double EARTH_RADIUS_MILES = 3956.0;

    /**
     * <p>The radius of the Earth, in kilometers.</p>
     */
    private static final double EARTH_RADIUS_KM = 6367.0;

    /**
     * <p>The 5-digit zip code that this ZipCode represents.</p>
     */
    private int _zipCode;

    /**
     * <p>The latitude, in degrees, of the post office serving this
     * ZipCode.</p>
     */
    private double _latitude;

    /**
     * <p>The longitude, in degrees, of the post office serving this
     * ZipCode.</p>
     */
    private double _longitude;

    /**
     * <p>Creates a new ZipCode object.</p>
     *
     * @param code The 5-digit zip code that the new ZipCode will represent.
     * @param latitude The latitude, in decimal degress, of the post office
     *        serving this zip code.
     * @param longitude The longitude, in decimal degress, of the post office
     *        serving this zip code.
     */
    public ZipCode(int code, double latitude, double longitude)
    {
	this._zipCode = code;
	this._latitude = latitude;
	this._longitude = longitude;
    }

    /**
     * <p>Gets the 5-digit zip code for this ZipCode.</p>
     *
     * @return The 5-digit zip code for this ZipCode.
     */
    public int getZipCode()
    {
	return _zipCode;
    }

    /**
     * <p>Gets the latitude of the post office serving this zip code.</p>
     *
     * @return The latitude of the post office serving this zip code.
     */
    public double getLatitude()
    {
	return _latitude;
    }

    /**
     * <p>Gets the longitude of the post office serving this zip code.</p>
     *
     * @return The longitude of the post office serving this zip code.
     */
    public double getLongitude()
    {
	return _longitude;
    }

    /**
     * <p>Gets the distance, in miles, from this ZipCode to the specified
     * ZipCode.</p>
     *
     * <p>This method uses the {@link #getArcLength} method to calculate the
     * arc distance between the latitudes and longitudes for the two zip codes,
     * then multiplies by the radius of the Earth in miles.</p>
     *
     * @param zip The ZipCode for which the distance to this one should be
     *            calculated.
     *
     * @return The distance, in miles, from this ZipCode to the specified
     *         ZipCode.
     *
     * @see #getArcLength
     */
    public double getDistanceInMiles(ZipCode zip)
    {
	return EARTH_RADIUS_MILES * getArcLength(zip);
    }

    /**
     * <p>Gets the distance, in kilometers, from this ZipCode to the specified
     * ZipCode.</p>
     *
     * <p>This method uses the {@link #getArcLength} method to calculate the
     * arc distance between the latitudes and longitudes for the two zip codes,
     * then multiplies by the radius of the Earth in kilometers.</p>
     *
     * @param zip The ZipCode for which the distance to this one should be
     *            calculated.
     *
     * @return The distance, in kilometers, from this ZipCode to the specified
     *         ZipCode.
     *
     * @see #getArcLength
     */
    public double getDistanceInKilometers(ZipCode zip)
    {
	return EARTH_RADIUS_KM * getArcLength(zip);
    }

    /**
     * <p>Gets the length of the great circle arc, in radians, between this
     * ZipCode and the specified ZipCode.<p>
     *
     * <p>This method uses the haversine formula, which is supposed to be more
     * accurate for small distances. Since this will likely be used for small
     * distances, this algorithm was chosen.</p>
     *
     * <p>References: <a href="http://mathforum.org/library/drmath/view/51879.html">http://mathforum.org/library/drmath/view/51879.html</a></p>
     *
     * @param zip The zip code for which the arc length should be calculated.
     *
     * @return The length of the great circle arc, in radians, between this
     *         ZipCode and the specified ZipCode.
     */
    public double getArcLength(ZipCode zip)
    {
	if (this == zip
	    || (zip._latitude == this._latitude
		&& zip._longitude == this._longitude))
	    return 0.0; // Duh

	return arc_length(this._latitude,
			  this._longitude,
			  zip._latitude,
			  zip._longitude);
    }

    /**
     * @see #getArcLength
     */
    public static double arc_length(double lat1,
				    double lon1,
				    double lat2,
				    double lon2)
    {
	// Convert lat/lon to radians

	double conversionFactor = (double)Math.PI / (double)180.0;

	double lat1_rad = lat1 * conversionFactor;
	double lon1_rad = lon1 * conversionFactor;
	double lat2_rad = lat2 * conversionFactor;
	double lon2_rad = lon2 * conversionFactor;


	double dlat = lat2_rad - lat1_rad;
	double dlon = lon2_rad - lon1_rad;

	double a =
	    sqr(Math.sin(dlat / 2.0)) +
	    Math.cos(lat1_rad) 
	    * Math.cos(lat2_rad)
	    * sqr(Math.sin(dlon / 2.0));

	double c = 2.0 * Math.atan2(Math.sqrt(a), Math.sqrt(1-a));

	return c;
    }

    /**
     * Squares the double argument.<p>
     *
     * @param d The value to be squared.
     *
     * @return The square of the value <code>d</code>.
     */
    private static double sqr(double d)
    {
	return d*d;
    }

    /**
     * <p>Returns a String representation of this ZipCode (i.e. "20005"
     * or "00621").</p>
     *
     * @return A String representation of this ZipCode (i.e. "20005"
     * or "00621").
     */
    public String toString()
    {
	if(_zipCode < 10)
	    return "0000" + String.valueOf(_zipCode);
	else if(_zipCode < 100)
	    return "000" + String.valueOf(_zipCode);
	else if(_zipCode < 1000)
	    return "00" + String.valueOf(_zipCode);
	else
	    return String.valueOf(_zipCode);
    }

    /**
     * <p>Compares the specified object to this one.</p>
     *
     * <p>This allows sorting based upon numeric value of the Zip Code.</p>
     *
     * @see java.lang.Comparable#compareTo
     */
    public int compareTo(Object o)
    {
	return this._zipCode - ((ZipCode)o)._zipCode;
    }

    /**
     * <p>Returns <code>true</code> if the specifed object is equal to this
     * one.</p>
     *
     * <p>Equality in this case is defined as the other object being a ZipCode
     * and having the same <code>zipCode</code> as this one.<p>
     *
     * @return <code>true</code> if the specified object is equal to this one,
     *         <code>false</code> otherwise.
     */
    public boolean equals(Object o)
    {
	return this._zipCode == ((ZipCode)o)._zipCode;
    }
}
